@@ -24,13 +24,14 @@
 #include "data/read.h"
 #include "data/write.h"
 #include "data/keylist.h"
-#include "data/rebalance.h"
-
-#include "sb/io.h"
+#include "data/reconcile.h"
 
 #include "init/error.h"
 #include "init/recovery.h"
 
+#include "sb/counters.h"
+#include "sb/io.h"
+
 #include "util/enumerated_ref.h"
 #include "util/util.h"
 
@@ -141,6 +142,10 @@ int bch2_stripe_validate(struct bch_fs *c, struct bkey_s_c k,
 			 "invalid csum granularity (%u >= 64)",
 			 s->csum_granularity_bits);
 
+	bkey_fsck_err_on(!s->sectors,
+			 c, stripe_sectors_zero,
+			 "invalid sectors zero");
+
 	ret = bch2_bkey_ptrs_validate(c, k, from);
 fsck_err:
 	return ret;
@@ -400,7 +405,7 @@ int bch2_trigger_stripe(struct btree_trans *trans,
 
 		struct gc_stripe *gc = NULL;
 		if (flags & BTREE_TRIGGER_gc) {
-			gc = genradix_ptr_alloc(&c->gc_stripes, idx, GFP_KERNEL);
+			gc = genradix_ptr_alloc(&c->ec.gc_stripes, idx, GFP_KERNEL);
 			if (!gc) {
 				bch_err(c, "error allocating memory for gc_stripes, idx %llu", idx);
 				return bch_err_throw(c, ENOMEM_mark_stripe);
@@ -614,7 +619,7 @@ static void ec_validate_checksums(struct bch_fs *c, struct ec_stripe_buf *buf)
 					bch2_csum_err_msg(&err, v->csum_type, want, got);
 					prt_printf(&err, "  for %ps at %u of\n  ", (void *) _RET_IP_, i);
 					bch2_bkey_val_to_text(&err, c, bkey_i_to_s_c(&buf->key));
-					bch_err_ratelimited(ca, "%s", err.buf);
+					bch_err_dev_ratelimited(ca, "%s", err.buf);
 
 					bch2_io_error(ca, BCH_MEMBER_ERROR_checksum);
 				}
@@ -749,7 +754,7 @@ static void ec_block_io(struct bch_fs *c, struct ec_stripe_buf *buf,
 						       nr_iovecs,
 						       opf,
 						       GFP_KERNEL,
-						       &c->ec_bioset),
+						       &c->ec.block_bioset),
 				      struct ec_bio, bio);
 
 		ec_bio->ca			= ca;
@@ -822,8 +827,7 @@ int bch2_ec_read_extent(struct btree_trans *trans, struct bch_read_bio *rbio,
 	if (ret)
 		return stripe_reconstruct_err(c, orig_k, "-ENOMEM");
 
-	struct closure cl;
-	closure_init_stack(&cl);
+	CLASS(closure_stack, cl)();
 
 	for (unsigned i = 0; i < v->nr_blocks; i++)
 		ec_block_io(c, buf, REQ_OP_READ, i, &cl);
@@ -848,8 +852,8 @@ int bch2_ec_read_extent(struct btree_trans *trans, struct bch_read_bio *rbio,
 
 static int __ec_stripe_mem_alloc(struct bch_fs *c, size_t idx, gfp_t gfp)
 {
-	if (c->gc_pos.phase != GC_PHASE_not_running &&
-	    !genradix_ptr_alloc(&c->gc_stripes, idx, gfp))
+	if (c->gc.pos.phase != GC_PHASE_not_running &&
+	    !genradix_ptr_alloc(&c->ec.gc_stripes, idx, gfp))
 		return bch_err_throw(c, ENOMEM_ec_stripe_mem_alloc);
 
 	return 0;
@@ -873,10 +877,10 @@ static int ec_stripe_mem_alloc(struct btree_trans *trans,
 
 static bool __bch2_bucket_has_new_stripe(struct bch_fs *c, u64 dev_bucket)
 {
-	unsigned hash = hash_64(dev_bucket, ilog2(ARRAY_SIZE(c->ec_stripes_new_buckets)));
+	unsigned hash = hash_64(dev_bucket, ilog2(ARRAY_SIZE(c->ec.stripes_new_buckets)));
 	struct ec_stripe_new_bucket *s;
 
-	hlist_for_each_entry(s, &c->ec_stripes_new_buckets[hash], hash)
+	hlist_for_each_entry(s, &c->ec.stripes_new_buckets[hash], hash)
 		if (s->dev_bucket == dev_bucket)
 			return true;
 	return false;
@@ -884,7 +888,7 @@ static bool __bch2_bucket_has_new_stripe(struct bch_fs *c, u64 dev_bucket)
 
 bool bch2_bucket_has_new_stripe(struct bch_fs *c, u64 dev_bucket)
 {
-	guard(spinlock)(&c->ec_stripes_new_lock);
+	guard(spinlock)(&c->ec.stripes_new_lock);
 	return __bch2_bucket_has_new_stripe(c, dev_bucket);
 }
 
@@ -892,20 +896,20 @@ static void stripe_new_bucket_add(struct bch_fs *c, struct ec_stripe_new_bucket
 {
 	s->dev_bucket = dev_bucket;
 
-	unsigned hash = hash_64(dev_bucket, ilog2(ARRAY_SIZE(c->ec_stripes_new_buckets)));
-	hlist_add_head(&s->hash, &c->ec_stripes_new_buckets[hash]);
+	unsigned hash = hash_64(dev_bucket, ilog2(ARRAY_SIZE(c->ec.stripes_new_buckets)));
+	hlist_add_head(&s->hash, &c->ec.stripes_new_buckets[hash]);
 }
 
 static void stripe_new_buckets_add(struct bch_fs *c, struct ec_stripe_new *s)
 {
 	unsigned nr_blocks = s->nr_data + s->nr_parity;
 
-	guard(spinlock)(&c->ec_stripes_new_lock);
+	guard(spinlock)(&c->ec.stripes_new_lock);
 	for (unsigned i = 0; i < nr_blocks; i++) {
 		if (!s->blocks[i])
 			continue;
 
-		struct open_bucket *ob = c->open_buckets + s->blocks[i];
+		struct open_bucket *ob = c->allocator.open_buckets + s->blocks[i];
 		struct bpos bucket = POS(ob->dev, ob->bucket);
 
 		stripe_new_bucket_add(c, &s->buckets[i], bucket_to_u64(bucket));
@@ -914,7 +918,7 @@ static void stripe_new_buckets_add(struct bch_fs *c, struct ec_stripe_new *s)
 
 static void stripe_new_buckets_del(struct bch_fs *c, struct ec_stripe_new *s)
 {
-	guard(spinlock)(&c->ec_stripes_new_lock);
+	guard(spinlock)(&c->ec.stripes_new_lock);
 
 	struct bch_stripe *v = &bkey_i_to_stripe(&s->new_stripe.key)->v;
 	for (unsigned i = 0; i < v->nr_blocks; i++)
@@ -923,10 +927,10 @@ static void stripe_new_buckets_del(struct bch_fs *c, struct ec_stripe_new *s)
 
 static struct ec_stripe_handle *bch2_open_stripe_find(struct bch_fs *c, u64 idx)
 {
-	unsigned hash = hash_64(idx, ilog2(ARRAY_SIZE(c->ec_stripes_new)));
+	unsigned hash = hash_64(idx, ilog2(ARRAY_SIZE(c->ec.stripes_new)));
 	struct ec_stripe_handle *s;
 
-	hlist_for_each_entry(s, &c->ec_stripes_new[hash], hash)
+	hlist_for_each_entry(s, &c->ec.stripes_new[hash], hash)
 		if (s->idx == idx)
 			return s;
 	return NULL;
@@ -934,7 +938,7 @@ static struct ec_stripe_handle *bch2_open_stripe_find(struct bch_fs *c, u64 idx)
 
 static bool bch2_stripe_is_open(struct bch_fs *c, u64 idx)
 {
-	guard(spinlock)(&c->ec_stripes_new_lock);
+	guard(spinlock)(&c->ec.stripes_new_lock);
 	return bch2_open_stripe_find(c, idx) != NULL;
 }
 
@@ -945,13 +949,13 @@ static bool bch2_stripe_handle_tryget(struct bch_fs *c,
 	BUG_ON(s->idx);
 	BUG_ON(!idx);
 
-	guard(spinlock)(&c->ec_stripes_new_lock);
+	guard(spinlock)(&c->ec.stripes_new_lock);
 	bool ret = !bch2_open_stripe_find(c, idx);
 	if (ret) {
-		unsigned hash = hash_64(idx, ilog2(ARRAY_SIZE(c->ec_stripes_new)));
+		unsigned hash = hash_64(idx, ilog2(ARRAY_SIZE(c->ec.stripes_new)));
 
 		s->idx = idx;
-		hlist_add_head(&s->hash, &c->ec_stripes_new[hash]);
+		hlist_add_head(&s->hash, &c->ec.stripes_new[hash]);
 	}
 	return ret;
 }
@@ -961,7 +965,7 @@ static void bch2_stripe_handle_put(struct bch_fs *c, struct ec_stripe_handle *s)
 	if (!s->idx)
 		return;
 
-	guard(spinlock)(&c->ec_stripes_new_lock);
+	guard(spinlock)(&c->ec.stripes_new_lock);
 	BUG_ON(bch2_open_stripe_find(c, s->idx) != s);
 	hlist_del_init(&s->hash);
 
@@ -994,7 +998,7 @@ static int ec_stripe_delete(struct btree_trans *trans, u64 idx)
 static void ec_stripe_delete_work(struct work_struct *work)
 {
 	struct bch_fs *c =
-		container_of(work, struct bch_fs, ec_stripe_delete_work);
+		container_of(work, struct bch_fs, ec.stripe_delete_work);
 
 	bch2_trans_run(c,
 		bch2_btree_write_buffer_tryflush(trans) ?:
@@ -1012,7 +1016,7 @@ static void ec_stripe_delete_work(struct work_struct *work)
 void bch2_do_stripe_deletes(struct bch_fs *c)
 {
 	if (enumerated_ref_tryget(&c->writes, BCH_WRITE_REF_stripe_delete) &&
-	    !queue_work(c->write_ref_wq, &c->ec_stripe_delete_work))
+	    !queue_work(c->write_ref_wq, &c->ec.stripe_delete_work))
 		enumerated_ref_put(&c->writes, BCH_WRITE_REF_stripe_delete);
 }
 
@@ -1091,6 +1095,7 @@ static int ec_stripe_update_extent(struct btree_trans *trans,
 				   struct ec_stripe_buf *s,
 				   struct bkey_s_c_backpointer bp,
 				   struct stripe_update_bucket_stats *stats,
+				   struct disk_reservation *res,
 				   struct wb_maybe_flush *last_flushed)
 {
 	struct bch_stripe *v = &bkey_i_to_stripe(&s->key)->v;
@@ -1121,7 +1126,12 @@ static int ec_stripe_update_extent(struct btree_trans *trans,
 		 */
 		stats->nr_bp_to_deleted++;
 		stats->sectors_bp_to_deleted += bp.v->bucket_len;
-		count_event(c, ec_stripe_update_extent_fail);
+		event_inc_trace(c, stripe_update_extent, buf, ({
+			prt_str(&buf, "backpointer race\n");
+			bch2_bkey_val_to_text(&buf, c, bp.s_c);
+			prt_newline(&buf);
+			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&s->key));
+		}));
 		return 0;
 	}
 
@@ -1137,13 +1147,23 @@ static int ec_stripe_update_extent(struct btree_trans *trans,
 	if (!ptr_c) {
 		stats->nr_no_match++;
 		stats->sectors_no_match += bp.v->bucket_len;
-		count_event(c, ec_stripe_update_extent_fail);
+		event_inc_trace(c, stripe_update_extent, buf, ({
+			prt_str(&buf, "no matching pointer found\n");
+			bch2_bkey_val_to_text(&buf, c, k);
+			prt_newline(&buf);
+			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&s->key));
+		}));
 		return 0;
 	}
 	if (ptr_c->cached) {
 		stats->nr_cached++;
 		stats->sectors_cached += bp.v->bucket_len;
-		count_event(c, ec_stripe_update_extent_fail);
+		event_inc_trace(c, stripe_update_extent, buf, ({
+			prt_str(&buf, "cached pointer\n");
+			bch2_bkey_val_to_text(&buf, c, k);
+			prt_newline(&buf);
+			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&s->key));
+		}));
 		return 0;
 	}
 
@@ -1156,7 +1176,7 @@ static int ec_stripe_update_extent(struct btree_trans *trans,
 		.idx		= s->key.k.p.offset,
 	};
 
-	struct bkey_i *n = errptr_try(bch2_trans_kmalloc(trans, bkey_bytes(k.k) + sizeof(stripe_ptr)));
+	struct bkey_i *n = errptr_try(bch2_trans_kmalloc(trans, BKEY_EXTENT_U64s_MAX * sizeof(u64)));
 
 	bkey_reassemble(n, k);
 
@@ -1172,17 +1192,17 @@ static int ec_stripe_update_extent(struct btree_trans *trans,
 	struct bch_inode_opts opts;
 
 	try(bch2_bkey_get_io_opts(trans, NULL, bkey_i_to_s_c(n), &opts));
-	try(bch2_bkey_set_needs_rebalance(trans->c, &opts, n,
-					  SET_NEEDS_REBALANCE_other, 0));
+	try(bch2_bkey_set_needs_reconcile(trans, NULL, &opts, n, SET_NEEDS_REBALANCE_other, 0));
 	try(bch2_trans_update(trans, &iter, n, 0));
-	try(bch2_trans_commit(trans, NULL, NULL,
+	try(bch2_trans_commit(trans, res, NULL,
 			BCH_TRANS_COMMIT_no_check_rw|
 			BCH_TRANS_COMMIT_no_enospc));
 
 	stats->nr_done++;
 	stats->sectors_done += bp.v->bucket_len;
 
-	count_event(c, ec_stripe_update_extent);
+	event_inc_trace(c, stripe_update_extent, buf,
+		bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(n)));
 
 	return 0;
 }
@@ -1205,6 +1225,8 @@ static int ec_stripe_update_bucket(struct btree_trans *trans, struct ec_stripe_b
 
 	struct stripe_update_bucket_stats stats = {};
 
+	CLASS(disk_reservation, res)(c);
+
 	try(for_each_btree_key_max(trans, bp_iter, BTREE_ID_backpointers,
 			bucket_pos_to_bp_start(ca, bucket_pos),
 			bucket_pos_to_bp_end(ca, bucket_pos), 0, bp_k, ({
@@ -1220,12 +1242,10 @@ static int ec_stripe_update_bucket(struct btree_trans *trans, struct ec_stripe_b
 
 		wb_maybe_flush_inc(&last_flushed);
 		ec_stripe_update_extent(trans, ca, bucket_pos, ptr.gen, s, bp,
-					&stats, &last_flushed);
+					&stats, &res.r, &last_flushed);
 	})));
 
-	if (trace_stripe_update_bucket_enabled()) {
-		CLASS(printbuf, buf)();
-
+	event_inc_trace(c, stripe_update_bucket, buf, ({
 		prt_printf(&buf, "bp_to_deleted:\t%u %u\n",
 			   stats.nr_bp_to_deleted, stats.sectors_bp_to_deleted);
 		prt_printf(&buf, "no_match:\t%u %u\n",
@@ -1234,9 +1254,7 @@ static int ec_stripe_update_bucket(struct btree_trans *trans, struct ec_stripe_b
 			   stats.nr_cached, stats.sectors_cached);
 		prt_printf(&buf, "done:\t%u %u\n",
 			   stats.nr_done, stats.sectors_done);
-
-		trace_stripe_update_bucket(c, buf.buf);
-	}
+	}));
 
 	return 0;
 }
@@ -1305,7 +1323,7 @@ static int __ec_stripe_create(struct ec_stripe_new *s)
 
 	for (unsigned i = 0; i < nr_data; i++)
 		if (s->blocks[i]) {
-			struct open_bucket *ob = c->open_buckets + s->blocks[i];
+			struct open_bucket *ob = c->allocator.open_buckets + s->blocks[i];
 
 			if (ob->sectors_free)
 				zero_out_rest_of_ec_bucket(c, s, i, ob);
@@ -1372,13 +1390,20 @@ static void ec_stripe_create(struct ec_stripe_new *s)
 	if (ret && !s->err)
 		s->err = ret;
 
-	trace_stripe_create(c, s->new_stripe.key.k.p.offset, ret);
+	if (!ret)
+		event_inc_trace(c, stripe_create, buf,
+			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&s->new_stripe.key)));
+	else
+		event_inc_trace(c, stripe_create_fail, buf, ({
+			prt_printf(&buf, "error %s\n", bch2_err_str(ret));
+			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(&s->new_stripe.key));
+		}));
 
 	bch2_disk_reservation_put(c, &s->res);
 
 	for (unsigned i = 0; i < v->nr_blocks; i++)
 		if (s->blocks[i]) {
-			struct open_bucket *ob = c->open_buckets + s->blocks[i];
+			struct open_bucket *ob = c->allocator.open_buckets + s->blocks[i];
 
 			if (i < nr_data) {
 				ob->ec = NULL;
@@ -1388,9 +1413,9 @@ static void ec_stripe_create(struct ec_stripe_new *s)
 			}
 		}
 
-	scoped_guard(mutex, &c->ec_stripe_new_lock)
+	scoped_guard(mutex, &c->ec.stripe_new_lock)
 		list_del(&s->list);
-	wake_up(&c->ec_stripe_new_wait);
+	wake_up(&c->ec.stripe_new_wait);
 
 	ec_stripe_buf_exit(&s->old_stripe);
 	ec_stripe_buf_exit(&s->new_stripe);
@@ -1403,8 +1428,8 @@ static struct ec_stripe_new *get_pending_stripe(struct bch_fs *c)
 {
 	struct ec_stripe_new *s;
 
-	guard(mutex)(&c->ec_stripe_new_lock);
-	list_for_each_entry(s, &c->ec_stripe_new_list, list)
+	guard(mutex)(&c->ec.stripe_new_lock);
+	list_for_each_entry(s, &c->ec.stripe_new_list, list)
 		if (!atomic_read(&s->ref[STRIPE_REF_io]))
 			return s;
 	return NULL;
@@ -1413,7 +1438,7 @@ static struct ec_stripe_new *get_pending_stripe(struct bch_fs *c)
 static void ec_stripe_create_work(struct work_struct *work)
 {
 	struct bch_fs *c = container_of(work,
-		struct bch_fs, ec_stripe_create_work);
+		struct bch_fs, ec.stripe_create_work);
 	struct ec_stripe_new *s;
 
 	while ((s = get_pending_stripe(c)))
@@ -1426,7 +1451,7 @@ void bch2_ec_do_stripe_creates(struct bch_fs *c)
 {
 	enumerated_ref_get(&c->writes, BCH_WRITE_REF_stripe_create);
 
-	if (!queue_work(system_long_wq, &c->ec_stripe_create_work))
+	if (!queue_work(system_long_wq, &c->ec.stripe_create_work))
 		enumerated_ref_put(&c->writes, BCH_WRITE_REF_stripe_create);
 }
 
@@ -1441,8 +1466,8 @@ static void ec_stripe_new_set_pending(struct bch_fs *c, struct ec_stripe_head *h
 	h->s		= NULL;
 	s->pending	= true;
 
-	scoped_guard(mutex, &c->ec_stripe_new_lock)
-		list_add(&s->list, &c->ec_stripe_new_list);
+	scoped_guard(mutex, &c->ec.stripe_new_lock)
+		list_add(&s->list, &c->ec.stripe_new_list);
 
 	ec_stripe_new_put(c, s, STRIPE_REF_io);
 }
@@ -1624,7 +1649,7 @@ static void ec_stripe_head_devs_update(struct bch_fs *c, struct ec_stripe_head *
 	if (h->s && !h->s->allocated && dev_mask_nr(&devs_leaving))
 		ec_stripe_new_cancel(c, h, -EINTR);
 
-	h->rw_devs_change_count = c->rw_devs_change_count;
+	h->rw_devs_change_count = c->allocator.rw_devs_change_count;
 }
 
 static struct ec_stripe_head *
@@ -1646,7 +1671,7 @@ ec_new_stripe_head_alloc(struct bch_fs *c, unsigned disk_label,
 	h->redundancy	= redundancy;
 	h->watermark	= watermark;
 
-	list_add(&h->list, &c->ec_stripe_head_list);
+	list_add(&h->list, &c->ec.stripe_head_list);
 	return h;
 }
 
@@ -1674,7 +1699,7 @@ __bch2_ec_stripe_head_get(struct btree_trans *trans,
 	if (!redundancy)
 		return NULL;
 
-	int ret = bch2_trans_mutex_lock(trans, &c->ec_stripe_head_lock);
+	int ret = bch2_trans_mutex_lock(trans, &c->ec.stripe_head_lock);
 	if (ret)
 		return ERR_PTR(ret);
 
@@ -1683,7 +1708,7 @@ __bch2_ec_stripe_head_get(struct btree_trans *trans,
 		goto err;
 	}
 
-	list_for_each_entry(h, &c->ec_stripe_head_list, list)
+	list_for_each_entry(h, &c->ec.stripe_head_list, list)
 		if (h->disk_label	== disk_label &&
 		    h->algo		== algo &&
 		    h->redundancy	== redundancy &&
@@ -1702,7 +1727,7 @@ __bch2_ec_stripe_head_get(struct btree_trans *trans,
 		goto err;
 	}
 found:
-	if (h->rw_devs_change_count != c->rw_devs_change_count)
+	if (h->rw_devs_change_count != c->allocator.rw_devs_change_count)
 		ec_stripe_head_devs_update(c, h);
 
 	if (h->insufficient_devs) {
@@ -1710,7 +1735,7 @@ found:
 		h = NULL;
 	}
 err:
-	mutex_unlock(&c->ec_stripe_head_lock);
+	mutex_unlock(&c->ec.stripe_head_lock);
 	return h;
 }
 
@@ -1729,7 +1754,7 @@ static int __new_stripe_alloc_buckets(struct btree_trans *trans,
 
 	/* * We bypass the sector allocator which normally does this: */
 	bitmap_and(req->devs_may_alloc.d, req->devs_may_alloc.d,
-		   c->rw_devs[BCH_DATA_user].d, BCH_SB_MEMBERS_MAX);
+		   c->allocator.rw_devs[BCH_DATA_user].d, BCH_SB_MEMBERS_MAX);
 
 	for_each_set_bit(i, s->blocks_gotten, v->nr_blocks) {
 		/*
@@ -1882,7 +1907,7 @@ static int init_new_stripe_from_old(struct bch_fs *c, struct ec_stripe_new *s)
 	 * blocks from the stripe we're reusing:
 	 */
 	for_each_set_bit(i, s->blocks_gotten, new_v->nr_blocks) {
-		bch2_open_bucket_put(c, c->open_buckets + s->blocks[i]);
+		bch2_open_bucket_put(c, c->allocator.open_buckets + s->blocks[i]);
 		s->blocks[i] = 0;
 	}
 	memset(s->blocks_gotten, 0, sizeof(s->blocks_gotten));
@@ -1945,12 +1970,12 @@ static int stripe_idx_alloc(struct btree_trans *trans, struct ec_stripe_new *s)
 	struct bch_fs *c = trans->c;
 	struct bkey_s_c k;
 	struct bpos min_pos = POS(0, 1);
-	struct bpos start_pos = bpos_max(min_pos, POS(0, c->ec_stripe_hint));
+	struct bpos start_pos = bpos_max(min_pos, POS(0, c->ec.stripe_hint));
 	int ret;
 
 	for_each_btree_key_norestart(trans, iter, BTREE_ID_stripes, start_pos,
 			   BTREE_ITER_slots|BTREE_ITER_intent, k, ret) {
-		c->ec_stripe_hint = iter.pos.offset;
+		c->ec.stripe_hint = iter.pos.offset;
 
 		if (bkey_gt(k.k->p, POS(0, U32_MAX))) {
 			if (start_pos.offset) {
@@ -2016,7 +2041,7 @@ static int stripe_alloc_or_reuse(struct btree_trans *trans,
 				}
 
 				/* XXX freelist_wait? */
-				closure_wait(&c->freelist_wait, cl);
+				closure_wait(&c->allocator.freelist_wait, cl);
 				*waiting = true;
 			}
 		}
@@ -2082,7 +2107,7 @@ struct ec_stripe_head *bch2_ec_stripe_head_get(struct btree_trans *trans,
 		ret = stripe_alloc_or_reuse(trans, req, cl, h, s, &waiting);
 		if (waiting &&
 		    !bch2_err_matches(ret, BCH_ERR_operation_blocked))
-			closure_wake_up(&c->freelist_wait);
+			closure_wake_up(&c->allocator.freelist_wait);
 
 		if (ret)
 			goto err;
@@ -2134,7 +2159,7 @@ int bch2_invalidate_stripe_to_dev(struct btree_trans *trans,
 				ptr->dev = BCH_SB_MEMBER_INVALID;
 
 			struct bch_dev *ca = bch2_dev_rcu(c, ptr->dev);
-			nr_good += ca && ca->mi.state != BCH_MEMBER_STATE_failed;
+			nr_good += ca && ca->mi.state != BCH_MEMBER_STATE_evacuating;
 		}
 
 	if (nr_good < s->v.nr_blocks && !(flags & BCH_FORCE_IF_DATA_DEGRADED)) {
@@ -2208,7 +2233,7 @@ static bool should_cancel_stripe(struct bch_fs *c, struct ec_stripe_new *s, stru
 		if (!s->blocks[i])
 			continue;
 
-		struct open_bucket *ob = c->open_buckets + s->blocks[i];
+		struct open_bucket *ob = c->allocator.open_buckets + s->blocks[i];
 		if (ob->dev == ca->dev_idx)
 			return true;
 	}
@@ -2220,8 +2245,8 @@ static void __bch2_ec_stop(struct bch_fs *c, struct bch_dev *ca)
 {
 	struct ec_stripe_head *h;
 
-	guard(mutex)(&c->ec_stripe_head_lock);
-	list_for_each_entry(h, &c->ec_stripe_head_list, list) {
+	guard(mutex)(&c->ec.stripe_head_lock);
+	list_for_each_entry(h, &c->ec.stripe_head_list, list) {
 		guard(mutex)(&h->lock);
 		if (h->s && should_cancel_stripe(c, h->s, ca))
 			ec_stripe_new_cancel(c, h, -BCH_ERR_erofs_no_writes);
@@ -2242,13 +2267,13 @@ static bool bch2_fs_ec_flush_done(struct bch_fs *c)
 {
 	sched_annotate_sleep();
 
-	guard(mutex)(&c->ec_stripe_new_lock);
-	return list_empty(&c->ec_stripe_new_list);
+	guard(mutex)(&c->ec.stripe_new_lock);
+	return list_empty(&c->ec.stripe_new_list);
 }
 
 void bch2_fs_ec_flush(struct bch_fs *c)
 {
-	wait_event(c->ec_stripe_new_wait, bch2_fs_ec_flush_done(c));
+	wait_event(c->ec.stripe_new_wait, bch2_fs_ec_flush_done(c));
 }
 
 int bch2_stripes_read(struct bch_fs *c)
@@ -2280,8 +2305,8 @@ void bch2_new_stripes_to_text(struct printbuf *out, struct bch_fs *c)
 	struct ec_stripe_head *h;
 	struct ec_stripe_new *s;
 
-	scoped_guard(mutex, &c->ec_stripe_head_lock)
-		list_for_each_entry(h, &c->ec_stripe_head_list, list) {
+	scoped_guard(mutex, &c->ec.stripe_head_lock)
+		list_for_each_entry(h, &c->ec.stripe_head_list, list) {
 			prt_printf(out, "disk label %u algo %u redundancy %u %s nr created %llu:\n",
 			       h->disk_label, h->algo, h->redundancy,
 			       bch2_watermarks[h->watermark],
@@ -2293,8 +2318,8 @@ void bch2_new_stripes_to_text(struct printbuf *out, struct bch_fs *c)
 
 	prt_printf(out, "in flight:\n");
 
-	scoped_guard(mutex, &c->ec_stripe_new_lock)
-		list_for_each_entry(s, &c->ec_stripe_new_list, list)
+	scoped_guard(mutex, &c->ec.stripe_new_lock)
+		list_for_each_entry(s, &c->ec.stripe_new_list, list)
 			bch2_new_stripe_to_text(out, c, s);
 }
 
@@ -2304,8 +2329,8 @@ void bch2_fs_ec_exit(struct bch_fs *c)
 	while (1) {
 		struct ec_stripe_head *h;
 
-		scoped_guard(mutex, &c->ec_stripe_head_lock)
-			h = list_pop_entry(&c->ec_stripe_head_list, struct ec_stripe_head, list);
+		scoped_guard(mutex, &c->ec.stripe_head_lock)
+			h = list_pop_entry(&c->ec.stripe_head_list, struct ec_stripe_head, list);
 
 		if (!h)
 			break;
@@ -2321,29 +2346,29 @@ void bch2_fs_ec_exit(struct bch_fs *c)
 		kfree(h);
 	}
 
-	BUG_ON(!list_empty(&c->ec_stripe_new_list));
+	BUG_ON(!list_empty(&c->ec.stripe_new_list));
 
-	bioset_exit(&c->ec_bioset);
+	bioset_exit(&c->ec.block_bioset);
 }
 
 void bch2_fs_ec_init_early(struct bch_fs *c)
 {
-	spin_lock_init(&c->ec_stripes_new_lock);
+	spin_lock_init(&c->ec.stripes_new_lock);
 
-	INIT_LIST_HEAD(&c->ec_stripe_head_list);
-	mutex_init(&c->ec_stripe_head_lock);
+	INIT_LIST_HEAD(&c->ec.stripe_head_list);
+	mutex_init(&c->ec.stripe_head_lock);
 
-	INIT_LIST_HEAD(&c->ec_stripe_new_list);
-	mutex_init(&c->ec_stripe_new_lock);
-	init_waitqueue_head(&c->ec_stripe_new_wait);
+	INIT_LIST_HEAD(&c->ec.stripe_new_list);
+	mutex_init(&c->ec.stripe_new_lock);
+	init_waitqueue_head(&c->ec.stripe_new_wait);
 
-	INIT_WORK(&c->ec_stripe_create_work, ec_stripe_create_work);
-	INIT_WORK(&c->ec_stripe_delete_work, ec_stripe_delete_work);
+	INIT_WORK(&c->ec.stripe_create_work, ec_stripe_create_work);
+	INIT_WORK(&c->ec.stripe_delete_work, ec_stripe_delete_work);
 }
 
 int bch2_fs_ec_init(struct bch_fs *c)
 {
-	return bioset_init(&c->ec_bioset, 1, offsetof(struct ec_bio, bio),
+	return bioset_init(&c->ec.block_bioset, 1, offsetof(struct ec_bio, bio),
 			   BIOSET_NEED_BVECS);
 }
 
