@@ -15,7 +15,7 @@
 #include "data/keylist.h"
 #include "data/move.h"
 #include "data/nocow_locking.h"
-#include "data/rebalance.h"
+#include "data/reconcile.h"
 #include "data/update.h"
 #include "data/write.h"
 
@@ -24,6 +24,8 @@
 #include "init/error.h"
 #include "init/fs.h"
 
+#include "sb/counters.h"
+
 #include "snapshots/snapshot.h"
 #include "snapshots/subvolume.h"
 
@@ -63,88 +65,64 @@ static unsigned bkey_get_dev_refs(struct bch_fs *c, struct bkey_s_c k)
 }
 
 noinline_for_stack
-static void trace_data_update_key_fail2(struct data_update *m,
-					struct btree_iter *iter,
-					struct bkey_s_c new,
-					struct bkey_s_c wrote,
-					struct bkey_i *insert,
-					const char *msg)
+static void count_data_update_key_fail(struct data_update *u,
+				       struct bkey_s_c new,
+				       struct bkey_s_c wrote,
+				       struct bkey_i *insert,
+				       const char *msg)
 {
-	if (m->stats) {
-		atomic64_inc(&m->stats->keys_raced);
-		atomic64_add(new.k->p.offset - iter->pos.offset,
-			     &m->stats->sectors_raced);
+	struct bch_fs *c = u->op.c;
+
+	if (u->stats) {
+		atomic64_inc(&u->stats->keys_raced);
+		atomic64_add(insert->k.size, &u->stats->sectors_raced);
 	}
 
-	count_event(m->op.c, data_update_key_fail);
+	event_add_trace(c, data_update_key_fail, insert->k.size, buf, ({
+		prt_str(&buf, bch2_data_update_type_strs[u->opts.type]);
+		prt_newline(&buf);
 
-	if (!trace_data_update_key_fail_enabled())
-		return;
+		prt_str(&buf, msg);
+		prt_newline(&buf);
 
-	struct bch_fs *c = m->op.c;
-	struct bkey_s_c old = bkey_i_to_s_c(m->k.k);
-	unsigned rewrites_found = 0;
+		struct bkey_s_c old = bkey_i_to_s_c(u->k.k);
+		unsigned rewrites_found = 0;
 
-	CLASS(printbuf, buf)();
-	printbuf_indent_add_nextline(&buf, 2);
+		if (insert) {
+			const union bch_extent_entry *entry;
+			struct bch_extent_ptr *ptr;
+			struct extent_ptr_decoded p;
 
-	prt_str(&buf, msg);
-	prt_newline(&buf);
-
-	if (insert) {
-		const union bch_extent_entry *entry;
-		struct bch_extent_ptr *ptr;
-		struct extent_ptr_decoded p;
-
-		unsigned ptr_bit = 1;
-		bkey_for_each_ptr_decode(old.k, bch2_bkey_ptrs_c(old), p, entry) {
-			if ((ptr_bit & m->opts.ptrs_rewrite) &&
-			    (ptr = bch2_extent_has_ptr(c, old, p, bkey_i_to_s(insert))) &&
-			    !ptr->cached)
-				rewrites_found |= ptr_bit;
-			ptr_bit <<= 1;
+			unsigned ptr_bit = 1;
+			bkey_for_each_ptr_decode(old.k, bch2_bkey_ptrs_c(old), p, entry) {
+				if ((ptr_bit & u->opts.ptrs_rewrite) &&
+				    (ptr = bch2_extent_has_ptr(c, old, p, bkey_i_to_s(insert))) &&
+				    !ptr->cached)
+					rewrites_found |= ptr_bit;
+				ptr_bit <<= 1;
+			}
 		}
-	}
 
-	prt_str(&buf, "rewrites found:\t");
-	bch2_prt_u64_base2(&buf, rewrites_found);
-	prt_newline(&buf);
+		prt_str(&buf, "rewrites found:\t");
+		bch2_prt_u64_base2(&buf, rewrites_found);
+		prt_newline(&buf);
 
-	bch2_data_update_opts_to_text(&buf, c, &m->op.opts, &m->opts);
+		bch2_data_update_opts_to_text(&buf, c, &u->op.opts, &u->opts);
 
-	prt_str_indented(&buf, "\nold:    ");
-	bch2_bkey_val_to_text(&buf, c, old);
+		prt_str_indented(&buf, "\nold:    ");
+		bch2_bkey_val_to_text(&buf, c, old);
 
-	prt_str_indented(&buf, "\nnew:    ");
-	bch2_bkey_val_to_text(&buf, c, new);
+		prt_str_indented(&buf, "\nnew:    ");
+		bch2_bkey_val_to_text(&buf, c, new);
 
-	prt_str_indented(&buf, "\nwrote:  ");
-	bch2_bkey_val_to_text(&buf, c, wrote);
+		prt_str_indented(&buf, "\nwrote:  ");
+		bch2_bkey_val_to_text(&buf, c, wrote);
 
-	if (insert) {
-		prt_str_indented(&buf, "\ninsert: ");
-		bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(insert));
-	}
-
-	trace_data_update_key_fail(c, buf.buf);
-}
-
-noinline_for_stack
-static void trace_data_update_key2(struct data_update *m,
-			       struct bkey_s_c old, struct bkey_s_c k,
-			       struct bkey_i *insert)
-{
-	struct bch_fs *c = m->op.c;
-	CLASS(printbuf, buf)();
-
-	prt_str(&buf, "\nold: ");
-	bch2_bkey_val_to_text(&buf, c, old);
-	prt_str(&buf, "\nk:   ");
-	bch2_bkey_val_to_text(&buf, c, k);
-	prt_str(&buf, "\nnew: ");
-	bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(insert));
-
-	trace_data_update_key(c, buf.buf);
+		if (insert) {
+			prt_str_indented(&buf, "\ninsert: ");
+			bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(insert));
+		}
+	}));
 }
 
 static int data_update_index_update_key(struct btree_trans *trans,
@@ -161,18 +139,32 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	/* make a local copy, so that we can trace it after the transaction commit:  */
 	k = bkey_i_to_s_c(errptr_try(bch2_bkey_make_mut_noupdate(trans, k)));
 
+	/*
+	 * We're calling set_needs_reconcile() on both @insert and @new,
+	 * and it can add a bch_extent_reconcile and additional
+	 * pointers to BCH_SB_MEMBER_INVALID if the extent is now
+	 * degraded due to option changes:
+	 */
 	struct bkey_i_extent *new = bkey_i_to_extent(bch2_keylist_front(&u->op.insert_keys));
-	new = errptr_try(bch2_trans_kmalloc(trans, bkey_bytes(&new->k)));
+	new = errptr_try(bch2_trans_kmalloc(trans, bkey_bytes(&new->k) +
+				 sizeof(struct bch_extent_reconcile) +
+				 sizeof(struct bch_extent_ptr) * BCH_REPLICAS_MAX));
 	bkey_copy(&new->k_i, bch2_keylist_front(&u->op.insert_keys));
 
 	struct bkey_i *insert = errptr_try(bch2_trans_kmalloc(trans,
 				    bkey_bytes(k.k) +
 				    bkey_val_bytes(&new->k) +
-				    sizeof(struct bch_extent_rebalance)));
+				    sizeof(struct bch_extent_reconcile) +
+				    sizeof(struct bch_extent_ptr) * BCH_REPLICAS_MAX));
 	bkey_reassemble(insert, k);
 
+	bch2_cut_front(c, iter->pos,	&new->k_i);
+	bch2_cut_front(c, iter->pos,	insert);
+	bch2_cut_back(new->k.p,		insert);
+	bch2_cut_back(insert->k.p,	&new->k_i);
+
 	if (!bch2_extents_match(c, k, old)) {
-		trace_data_update_key_fail2(u, iter, k, bkey_i_to_s_c(&new->k_i), NULL, "no match:");
+		count_data_update_key_fail(u, k, bkey_i_to_s_c(&new->k_i), insert, "no match:");
 		bch2_btree_iter_advance(iter);
 		return 0;
 	}
@@ -180,12 +172,6 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	struct bch_inode_opts opts;
 	try(bch2_bkey_get_io_opts(trans, NULL, k, &opts));
 
-	bch2_cut_front(c, iter->pos, &new->k_i);
-
-	bch2_cut_front(c, iter->pos,	insert);
-	bch2_cut_back(new->k.p,		insert);
-	bch2_cut_back(insert->k.p,	&new->k_i);
-
 	bch2_bkey_propagate_incompressible(c, insert, bkey_i_to_s_c(&new->k_i));
 
 	/*
@@ -218,8 +204,8 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	if (u->opts.ptrs_rewrite &&
 	    !rewrites_found &&
 	    bch2_bkey_durability(c, k) >= opts.data_replicas) {
-		trace_data_update_key_fail2(u, iter, k, bkey_i_to_s_c(&new->k_i), insert,
-					    "no rewrites found:");
+		count_data_update_key_fail(u, k, bkey_i_to_s_c(&new->k_i), insert,
+					   "no rewrites found:");
 		bch2_btree_iter_advance(iter);
 		return 0;
 	}
@@ -234,7 +220,7 @@ static int data_update_index_update_key(struct btree_trans *trans,
 		 !ptr_c->cached));
 
 	if (!bkey_val_u64s(&new->k)) {
-		trace_data_update_key_fail2(u, iter, k,
+		count_data_update_key_fail(u, k,
 				    bkey_i_to_s_c(bch2_keylist_front(&u->op.insert_keys)),
 				    insert, "new replicas conflicted:");
 		bch2_btree_iter_advance(iter);
@@ -274,7 +260,17 @@ static int data_update_index_update_key(struct btree_trans *trans,
 	try(bch2_insert_snapshot_whiteouts(trans, u->btree_id, k.k->p, bkey_start_pos(&insert->k)));
 	try(bch2_insert_snapshot_whiteouts(trans, u->btree_id, k.k->p, insert->k.p));
 
-	try(bch2_bkey_set_needs_rebalance(c, &opts, insert,
+	/*
+	 * This set_needs_reconcile call is only for verifying that the data we
+	 * just wrote was written correctly, otherwise we could fail to flag
+	 * incorrectly written data due to needs_rb already being set on the
+	 * existing extent
+	 */
+	try(bch2_bkey_set_needs_reconcile(trans, NULL, &opts, &new->k_i,
+					  SET_NEEDS_REBALANCE_foreground,
+					  u->op.opts.change_cookie));
+	/* This is the real set_needs_reconcile() call */
+	try(bch2_bkey_set_needs_reconcile(trans, NULL, &opts, insert,
 					  SET_NEEDS_REBALANCE_foreground,
 					  u->op.opts.change_cookie));
 
@@ -286,9 +282,15 @@ static int data_update_index_update_key(struct btree_trans *trans,
 
 	bch2_btree_iter_set_pos(iter, next_pos);
 
-	if (trace_data_update_key_enabled())
-		trace_data_update_key2(u, old, k, insert);
-	this_cpu_add(c->counters[BCH_COUNTER_data_update_key], new->k.size);
+	event_add_trace(c, data_update_key, new->k.size, buf, ({
+		prt_str(&buf, "\nold: ");
+		bch2_bkey_val_to_text(&buf, c, old);
+		prt_str(&buf, "\nk:   ");
+		bch2_bkey_val_to_text(&buf, c, k);
+		prt_str(&buf, "\nnew: ");
+		bch2_bkey_val_to_text(&buf, c, bkey_i_to_s_c(insert));
+	}));
+
 	return 0;
 }
 
@@ -347,6 +349,7 @@ void bch2_data_update_read_done(struct data_update *u)
 	}
 
 	if (unlikely(rbio->ret)) {
+		u->op.error = rbio->ret;
 		u->op.end_io(&u->op);
 		return;
 	}
@@ -385,35 +388,34 @@ void bch2_data_update_read_done(struct data_update *u)
 	closure_call(&u->op.cl, bch2_write, NULL, NULL);
 }
 
+static inline bool should_trace_update_err(struct data_update *u, int ret)
+{
+	if ((u->opts.type == BCH_DATA_UPDATE_reconcile ||
+	     u->opts.type == BCH_DATA_UPDATE_promote) &&
+	    (bch2_err_matches(ret, BCH_ERR_data_update_fail_no_rw_devs) ||
+	     bch2_err_matches(ret, BCH_ERR_insufficient_devices)))
+		return false;
+
+	return true;
+}
+
 static void data_update_trace(struct data_update *u, int ret)
 {
 	struct bch_fs *c = u->op.c;
 
-	if (!ret) {
-		if (trace_data_update_enabled()) {
-			CLASS(printbuf, buf)();
-			bch2_data_update_to_text(&buf, u);
-			trace_data_update(c, buf.buf);
-		}
-		count_event(c, data_update);
-	} else if (bch2_err_matches(ret, BCH_ERR_data_update_done)) {
-		if (trace_data_update_no_io_enabled()) {
-			CLASS(printbuf, buf)();
-			bch2_data_update_to_text(&buf, u);
-			prt_printf(&buf, "\nret:\t%s\n", bch2_err_str(ret));
-			trace_data_update_no_io(c, buf.buf);
-		}
-		count_event(c, data_update_no_io);
-	} else if (ret != -BCH_ERR_data_update_fail_no_rw_devs) {
-		if (trace_data_update_fail_enabled()) {
-			CLASS(printbuf, buf)();
-			bch2_data_update_to_text(&buf, u);
-			prt_printf(&buf, "\nret:\t%s\n", bch2_err_str(ret));
-			trace_data_update_fail(c, buf.buf);
-		}
-
-		count_event(c, data_update_fail);
-	}
+	if (!ret)
+		event_add_trace(c, data_update, u->k.k->k.size, buf,
+				bch2_data_update_to_text(&buf, u));
+	else if (bch2_err_matches(ret, BCH_ERR_data_update_done))
+		event_add_trace(c, data_update_no_io, u->k.k->k.size, buf, ({
+				bch2_data_update_to_text(&buf, u);
+				prt_printf(&buf, "\nret:\t%s\n", bch2_err_str(ret));
+		}));
+	else if (should_trace_update_err(u, ret))
+		event_add_trace(c, data_update_fail, u->k.k->k.size, buf, ({
+				bch2_data_update_to_text(&buf, u);
+				prt_printf(&buf, "\nret:\t%s\n", bch2_err_str(ret));
+		}));
 }
 
 void bch2_data_update_exit(struct data_update *update, int ret)
@@ -450,11 +452,10 @@ int bch2_update_unwritten_extent(struct btree_trans *trans,
 	struct bch_fs *c = update->op.c;
 	struct bkey_i_extent *e;
 	struct write_point *wp;
-	struct closure cl;
 	struct bkey_s_c k;
 	int ret = 0;
 
-	closure_init_stack(&cl);
+	CLASS(closure_stack, cl)();
 	bch2_keylist_init(&update->op.insert_keys, update->op.inline_keys);
 
 	while (bpos_lt(update->op.pos, update->k.k->k.p)) {
@@ -575,8 +576,10 @@ void bch2_data_update_to_text(struct printbuf *out, struct data_update *m)
 
 	prt_str_indented(out, "old key:\t");
 	bch2_bkey_val_to_text(out, m->op.c, bkey_i_to_s_c(m->k.k));
+	prt_newline(out);
 
-	bch2_write_op_to_text(out, &m->op);
+	prt_printf(out, "write: ");
+	__bch2_write_op_to_text(out, &m->op);
 }
 
 void bch2_data_update_inflight_to_text(struct printbuf *out, struct data_update *m)
@@ -593,7 +596,7 @@ void bch2_data_update_inflight_to_text(struct printbuf *out, struct data_update
 	} else {
 		prt_printf(out, "write:\n");
 		guard(printbuf_indent)(out);
-		bch2_write_op_to_text(out, &m->op);
+		__bch2_write_op_to_text(out, &m->op);
 	}
 }
 
@@ -607,21 +610,11 @@ static int bch2_extent_drop_ptrs(struct btree_trans *trans,
 
 	struct bkey_i *n = errptr_try(bch2_bkey_make_mut_noupdate(trans, k));
 
-	const union bch_extent_entry *entry;
-	struct extent_ptr_decoded p = {};
-	unsigned i = 0;
-	bkey_for_each_ptr_decode(k.k, bch2_bkey_ptrs_c(k), p, entry) {
-		if (data_opts->ptrs_kill_ec & BIT(i))
-			bch2_bkey_drop_ec(c, n, p.ptr.dev);
-		i++;
-	}
+	if (data_opts->ptrs_kill_ec)
+		bch2_bkey_drop_ec_mask(c, n, data_opts->ptrs_kill_ec);
 
-	while (data_opts->ptrs_kill) {
-		unsigned i = 0, drop = __fls(data_opts->ptrs_kill);
-
-		bch2_bkey_drop_ptrs_noerror(bkey_i_to_s(n), p, entry, i++ == drop);
-		data_opts->ptrs_kill ^= 1U << drop;
-	}
+	if (data_opts->ptrs_kill)
+		bch2_bkey_drop_ptrs_mask(c, n, data_opts->ptrs_kill);
 
 	/*
 	 * If the new extent no longer has any pointers, bch2_extent_normalize()
@@ -674,28 +667,37 @@ static int bch2_data_update_bios_init(struct data_update *m, struct bch_fs *c,
 	return 0;
 }
 
-static int can_write_extent(struct bch_fs *c, struct data_update *m)
+struct bch_devs_list bch2_data_update_devs_keeping(struct bch_fs *c,
+						   struct data_update_opts *opts,
+						   struct bkey_s_c k)
 {
-	if ((m->op.flags & BCH_WRITE_alloc_nowait) &&
-	    unlikely(c->open_buckets_nr_free <= bch2_open_buckets_reserved(m->op.watermark)))
-		return bch_err_throw(c, data_update_fail_would_block);
+	struct bch_devs_list ret = (struct bch_devs_list) { 0 };
 
-	unsigned target = m->op.flags & BCH_WRITE_only_specified_devs
-		? m->op.target
-		: 0;
+	/* We always rewrite btree nodes entirely */
+	if (bkey_is_btree_ptr(k.k))
+		return ret;
+
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	unsigned ptr_bit = 1;
+
+	bkey_for_each_ptr(ptrs, ptr) {
+		if (!(ptr_bit & (opts->ptrs_rewrite|
+				 opts->ptrs_kill)))
+			ret.data[ret.nr++] = ptr->dev;
+		ptr_bit <<= 1;
+	}
+
+	return ret;
+}
+
+static unsigned durability_available_on_target(struct bch_fs *c,
+					       enum bch_watermark watermark,
+					       unsigned target)
+{
 	struct bch_devs_mask devs = target_rw_devs(c, BCH_DATA_user, target);
+	unsigned ret = 0;
 
-	darray_for_each(m->op.devs_have, i)
-		if (*i != BCH_SB_MEMBER_INVALID)
-			__clear_bit(*i, devs.d);
-
-	bool trace = trace_data_update_fail_enabled();
-	CLASS(printbuf, buf)();
-
-	guard(printbuf_atomic)(&buf);
-	guard(rcu)();
-
-	unsigned nr_replicas = 0, i;
+	unsigned i;
 	for_each_set_bit(i, devs.d, BCH_SB_MEMBERS_MAX) {
 		struct bch_dev *ca = bch2_dev_rcu_noerror(c, i);
 		if (!ca)
@@ -704,39 +706,89 @@ static int can_write_extent(struct bch_fs *c, struct data_update *m)
 		struct bch_dev_usage usage;
 		bch2_dev_usage_read_fast(ca, &usage);
 
-		u64 nr_free = dev_buckets_free(ca, usage, m->op.watermark);
+		if (dev_buckets_free(ca, usage, watermark))
+			ret += ca->mi.durability;
+	}
 
-		if (trace)
-			prt_printf(&buf, "%s=%llu ", ca->name, nr_free);
+	return ret;
+}
 
-		if (!nr_free)
+static unsigned bch2_bkey_durability_on_target(struct bch_fs *c, struct bkey_s_c k,
+					       unsigned target)
+{
+	struct bch_devs_mask devs = target_rw_devs(c, BCH_DATA_user, target);
+	struct bkey_ptrs_c ptrs = bch2_bkey_ptrs_c(k);
+	const union bch_extent_entry *entry;
+	struct extent_ptr_decoded p;
+	unsigned durability = 0;
+
+	bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
+		if (p.ptr.dev != BCH_SB_MEMBER_INVALID &&
+		    test_bit(p.ptr.dev, devs.d))
+			durability += bch2_extent_ptr_durability(c, &p);
+	}
+	return durability;
+}
+
+static int bch2_can_do_write_btree(struct bch_fs *c,
+				   struct bch_inode_opts *opts,
+				   struct data_update_opts *data_opts, struct bkey_s_c k)
+{
+	enum bch_watermark watermark = data_opts->commit_flags & BCH_WATERMARK_MASK;
+
+	if (durability_available_on_target(c, watermark, data_opts->target) >
+	    bch2_bkey_durability_on_target(c, k, data_opts->target))
+		return 0;
+
+	if (!(data_opts->write_flags & BCH_WRITE_only_specified_devs)) {
+		unsigned d = bch2_bkey_durability(c, k);
+		if (d < opts->data_replicas &&
+		    d < durability_available_on_target(c, watermark, 0))
+			return 0;
+	}
+
+	return bch_err_throw(c, data_update_fail_no_rw_devs);
+}
+
+int bch2_can_do_write(struct bch_fs *c,
+		      struct bch_inode_opts *opts,
+		      struct data_update_opts *data_opts,
+		      struct bkey_s_c k, struct bch_devs_list *devs_have)
+{
+	enum bch_watermark watermark = data_opts->commit_flags & BCH_WATERMARK_MASK;
+
+	if ((data_opts->write_flags & BCH_WRITE_alloc_nowait) &&
+	    unlikely(c->allocator.open_buckets_nr_free <= bch2_open_buckets_reserved(watermark)))
+		return bch_err_throw(c, data_update_fail_would_block);
+
+	guard(rcu)();
+
+	if (bkey_is_btree_ptr(k.k))
+		return bch2_can_do_write_btree(c, opts, data_opts, k);
+
+	unsigned target = data_opts->write_flags & BCH_WRITE_only_specified_devs
+		? data_opts->target
+		: 0;
+	struct bch_devs_mask devs = target_rw_devs(c, BCH_DATA_user, target);
+
+	darray_for_each(*devs_have, i)
+		if (*i != BCH_SB_MEMBER_INVALID)
+			__clear_bit(*i, devs.d);
+
+	unsigned i;
+	for_each_set_bit(i, devs.d, BCH_SB_MEMBERS_MAX) {
+		struct bch_dev *ca = bch2_dev_rcu_noerror(c, i);
+		if (!ca)
 			continue;
 
-		nr_replicas += ca->mi.durability;
-		if (nr_replicas >= m->op.nr_replicas)
-			break;
+		struct bch_dev_usage usage;
+		bch2_dev_usage_read_fast(ca, &usage);
+
+		if (dev_buckets_free(ca, usage, watermark))
+			return 0;
 	}
 
-	if (!nr_replicas) {
-		/*
-		 * If it's a promote that's failing because the promote target
-		 * is full - we expect that in normal operation; it'll still
-		 * show up in io_read_nopromote and error_throw:
-		 */
-		if (m->opts.type != BCH_DATA_UPDATE_promote) {
-			if (trace) {
-				prt_printf(&buf, " - got replicas %u\n", nr_replicas);
-				bch2_data_update_to_text(&buf, m);
-				prt_printf(&buf, "\nret:\t%s\n", bch2_err_str(-BCH_ERR_data_update_fail_no_rw_devs));
-				trace_data_update_fail(c, buf.buf);
-			}
-			count_event(c, data_update_fail);
-		}
-
-		return bch_err_throw(c, data_update_fail_no_rw_devs);
-	}
-
-	return 0;
+	return bch_err_throw(c, data_update_fail_no_rw_devs);
 }
 
 /*
@@ -850,15 +902,18 @@ int bch2_data_update_init(struct btree_trans *trans,
 		unsigned ptr_bit = 1;
 		bkey_for_each_ptr_decode(k.k, ptrs, p, entry) {
 			if (!p.ptr.cached) {
+				if (!(ptr_bit & (m->opts.ptrs_rewrite|
+						 m->opts.ptrs_kill))) {
+					bch2_dev_list_add_dev(&m->op.devs_have, p.ptr.dev);
+					durability_have += bch2_extent_ptr_durability(c, &p);
+				}
+
 				if (ptr_bit & m->opts.ptrs_rewrite) {
 					if (crc_is_compressed(p.crc))
 						reserve_sectors += k.k->size;
 
 					m->op.nr_replicas += bch2_extent_ptr_desired_durability(c, &p);
 					durability_removing += bch2_extent_ptr_desired_durability(c, &p);
-				} else if (!(ptr_bit & m->opts.ptrs_kill)) {
-					bch2_dev_list_add_dev(&m->op.devs_have, p.ptr.dev);
-					durability_have += bch2_extent_ptr_durability(c, &p);
 				}
 			} else {
 				if (m->opts.ptrs_rewrite & ptr_bit) {
@@ -940,9 +995,11 @@ int bch2_data_update_init(struct btree_trans *trans,
 		 *   (i.e. trying to move a durability=2 replica to a target with a
 		 *   single durability=2 device)
 		 */
-		ret = can_write_extent(c, m);
-		if (ret)
-			goto out;
+		if (data_opts.type != BCH_DATA_UPDATE_copygc) {
+			ret = bch2_can_do_write(c, io_opts, &m->opts, k, &m->op.devs_have);
+			if (ret)
+				goto out;
+		}
 
 		if (reserve_sectors) {
 			ret = bch2_disk_reservation_add(c, &m->op.res, reserve_sectors,
@@ -1009,7 +1066,8 @@ out_nocow_unlock:
 out:
 	BUG_ON(!ret);
 
-	data_update_trace(m, ret);
+	if (!bch2_err_matches(ret, BCH_ERR_transaction_restart))
+		data_update_trace(m, ret);
 
 	bkey_put_dev_refs(c, k, m->ptrs_held);
 	m->ptrs_held = 0;
